local Lib = MissionMinder;
Lib.callbacks= Lib.callbacks or LibStub("CallbackHandler-1.0"):New(Lib)

local OPTIONS = MissionMinder.options;
local DEBUG = MissionMinder.debug;

MissionMinder.quickloot = {};
local module = MissionMinder.quickloot;

local L = LibStub("AceLocale-3.0"):GetLocale("MissionMinder");
local L_MISSION_SUCCESS = L["Mission %s was successful! Received %s"]
local L_MISSION_FAILED = L["Mission %s failed!"]
local L_FOLLOWER_XP_RECEIVED = L["Follower %s received +%d XP."]
local L_FOLLOWER_LEVELUP = L["Follower %s has reached level %d!"]
local L_FOLLOWER_QUALITYUP = L["Follower %s has reached %s quality!"]

local function createInlineTexture(texture)
	return texture and ("|T" .. texture .. ":0|t") or "";
end

local function getRewards(missionID)
	local rewards = C_Garrison.GetMissionRewardInfo(missionID);
	local str = {};

	for _, reward in pairs(rewards) do
		if(reward.followerXP) then
			local xp = BreakUpLargeNumbers(reward.followerXP)
			tinsert(str, "+" .. xp .. " " .. createInlineTexture(reward.icon));
		elseif (reward.itemID) then
			local _, link = GetItemInfo(reward.itemID);
			if(reward.quantity == 1) then
				tinsert(str, link);
			else
				tinsert(str, reward.quantity .. "x " .. link);
			end
		elseif(reward.currencyID) then
			if (reward.currencyID == 0) then -- gold
				tinsert(str, GetCoinTextureString(reward.quantity));
			else
				tinsert(str, reward.quantity .. " " .. createInlineTexture(reward.icon));
			end
		else
			tinsert(str, reward.title);
		end
	end

	return table.concat(str, ", ");
end

local function MissionHasBonusXPReward(missionID)
	local rewards = C_Garrison.GetMissionRewardInfo(missionID);
	for _, reward in pairs(rewards) do
		if(reward.followerXP) then
			return true;
		end
	end

	return false;
end

local lootedMissions = {} -- looted missions map is kept as long as inprogress is visible
local function missionSuccessFail(event, missionID, canComplete, succeeded)
	if ( OPTIONS.QuickLoot() ) then
		if ( lootedMissions[missionID] ) then
			return;
		else
			lootedMissions[missionID] = true
		end

		local missionLink = C_Garrison.GetMissionLink(missionID);
		if ( succeeded ) then
			local rewards = getRewards(missionID);
			SendSystemMessage(string.format(L_MISSION_SUCCESS, missionLink, rewards));
			PlaySound("UI_Garrison_CommandTable_MissionSuccess_Stinger");
		else
			SendSystemMessage(string.format(L_MISSION_FAILED, missionLink));
			PlaySound("UI_Garrison_Mission_Complete_MissionFail_Stinger");
		end

		if ( C_Garrison.CanOpenMissionChest(missionID)) then
			C_Garrison.MissionBonusRoll(missionID);
		end
	end
end

local function clearLootedMissions()
	lootedMissions = {}
end

local xpgains = {};
local function initializeXPState(event, missionID, canComplete, succeeded)
	local mission = Lib.findCompleteMission(missionID)
	local hasBonusXP = MissionHasBonusXPReward(missionID);

	for i=1, #(mission.followers) do
		local followerID = mission.followers[i];
		if ( succeeded and hasBonusXP ) then
			xpgains[followerID] = {};
			xpgains[followerID].bonusXP = true;
		elseif(xpgains[followerID]) then
			xpgains[followerID] = nil;
		-- else : if element is aleady nil, do nothing
		end
	end
end

local function reportXPGains(followerID, xpAward, oldXP, oldLevel, oldQuality)
	local follower =  C_Garrison.GetFollowerInfo(followerID);
	local followerLink = C_Garrison.GetFollowerLink(follower.followerID) or follower.name;

	local maxXP = GarrisonMissionFrame_GetFollowerNextLevelXP(oldLevel, oldQuality);
	if ( maxXP ) then  -- nil if already maxLevel and maxQuality

		if ( OPTIONS.ShowFollowerXPGains() ) then
			SendSystemMessage(string.format(L_FOLLOWER_XP_RECEIVED, followerLink, xpAward));
		end

		while ( maxXP and xpAward + oldXP >= maxXP ) do -- handle multiple level ups
			xpAward = xpAward - (maxXP - oldXP) -- calculate next level spillover
			oldXP = 0

			if ( oldLevel < GarrisonMissionFrame.followerMaxLevel ) then
				SendSystemMessage(string.format(L_FOLLOWER_LEVELUP, followerLink, oldLevel + 1));
				oldLevel = oldLevel + 1
				maxXP = GarrisonMissionFrame_GetFollowerNextLevelXP(oldLevel, oldQuality);
			elseif ( oldQuality < GarrisonMissionFrame.followerMaxQuality ) then
				SendSystemMessage(string.format(L_FOLLOWER_QUALITYUP, followerLink, getglobal("ITEM_QUALITY" .. oldQuality + 1 .. "_DESC")));
				oldQuality = oldQuality + 1
				maxXP = GarrisonMissionFrame_GetFollowerNextLevelXP(oldLevel, oldQuality);
			end
		end
	end
end

local function FollowerXPChanged(event, followerID, xpAward, oldXP, oldLevel, oldQuality)
	DEBUG("Follower %s received %d xp", followerID, xpAward);

	local gains = xpgains[followerID];
	if ( gains and gains.bonusXP ) then
		if ( not gains.xpAward ) then
			gains.xpAward = xpAward
			gains.oldXP = oldXP
			gains.oldLevel = oldLevel
			gains.oldQuality = oldQuality

			return -- bonusXP still expected, so we need to aggregate 2 separate GARRISON_FOLLOWER_XP_CHANGED events for each follower
		else
			xpAward = xpAward + gains.xpAward
			oldXP = gains.oldXP
			oldLevel = gains.oldLevel
			oldQuality = gains.oldQuality
		end
	end

	xpgains[followerID] = nil;

	reportXPGains(followerID, xpAward, oldXP, oldLevel, oldQuality)
end

local function initialize()
--function module.initialize()
	DEBUG("   |cffffffffQuickLoot|r initialized");

	local missionList = GarrisonMissionFrame.MissionTab.MissionList
	missionList:HookScript("OnHide", clearLootedMissions)
	missionList.Tab1:HookScript("OnClick", clearLootedMissions)
end

function module.lootMission(mission)
	if ( mission ) then
		if ( mission.state >= 0 ) then
			if ( C_Garrison.CanOpenMissionChest(mission.missionID) ) then
				C_Garrison.MissionBonusRoll(mission.missionID);
			else
				-- Mission cannot be looted; take no action
			end
		else
			C_Garrison.MarkMissionComplete(mission.missionID);
		end
	end
end

local events = Lib.events
events:RegisterEvent("quickloot", "ADDON_LOADED", initialize);
events:RegisterEvent("quickloot", "GARRISON_MISSION_COMPLETE_RESPONSE", missionSuccessFail);
events:RegisterEvent("quickloot_xp", "GARRISON_MISSION_COMPLETE_RESPONSE", initializeXPState);
events:RegisterEvent("quickloot", "GARRISON_FOLLOWER_XP_CHANGED", FollowerXPChanged);
